# QTFairUse6 2.4 2006-09-14
# Copyright (C) 2006 Igor Skochinsky <skochinsky@mail.ru>
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

from pydbg import *
from pydbg.defines import *
from ctypes import *
from dedrm import MP4, MP4Error
from iTunesController import iTunesController
import sys, struct, os, threading, os.path, urllib, re, platform, codecs, htmlentitydefs, time, inspect, Queue

error = lambda msg: sys.stderr.write("ERROR> " + msg + "\n") or sys.exit(1)

debugmode = False

def debug(msg):
    if debugmode:
        sys.stderr.write("[DEBUG] "+msg+"\n")
    pass

outf = None
fileno = 0
offset = 0
offsets1 = [0x4FB39A, 0x4FCB7A, 0x4D1615] #debug check
offsets2 = [0x976930, 0x989760, 0x7F18B0] #frame copy
offsets3 = [0x4b0ad0, 0x4B1180, 0x500C50] #end-of-track function
offsets4 = [0x77B929, 0x782459]           #SndDoCommand
versions = ["6.0.4", "6.0.5", "7.0.0"]

def find_pid(dbg, name):
    namel = name.lower()
    found_target = False
    for (pid, proc_name) in dbg.enumerate_processes():
        if proc_name.lower() == namel:
            return pid
    return -1

def getDwordAt(dbg, addr):
    return struct.unpack("<I", dbg.read_process_memory(addr, 4))[0]

def getWordAt(dbg, addr):
    return struct.unpack("<H", dbg.read_process_memory(addr, 2))[0]

def copy_process_memory (dbg, readaddress, writeaddress, length):
    """Copy data from debugee's readaddress to debugger's writeaddress"""
    count        = c_ulong(0)
    # ensure we can read from the requested memory space.
    _address = readaddress
    _length  = length

    try:
        old_protect = dbg.virtual_protect(_address, _length, PAGE_EXECUTE_READWRITE)
    except:
        pass

    while length:
        if not windll.kernel32.ReadProcessMemory(dbg.h_process, readaddress, writeaddress, length, byref(count)):
            raise pydbg.pdx("ReadProcessMemory(%08x, %d, read=%d)" % (address, length, count.value), True)

        #data    += read_buf.raw
        length  -= count.value
        readaddress += count.value
        writeaddress += count.value

    # restore the original page permissions on the target memory region.
    try:
        dbg.virtual_protect(_address, _length, old_protect)
    except:
        pass

    #return data

def bp_check (dbg):
    print "Hit %08X" % dbg.context.Eip
    return DBG_CONTINUE

def handleAV (dbg):
    print "Access violation at %08X (write: %d, address: %08X)" % (dbg.exception_address, dbg.write_violation, dbg.violation_address)
    return DBG_CONTINUE

class DebugThread(threading.Thread):

    def __init__(self):
        self.onstop = None
        self.debugger_attached = threading.Event()
        self.bufQueue = Queue.Queue(2)
        threading.Thread.__init__(self)

    def handler_breakpoint (self, dbg):
        if dbg.first_breakpoint:
            dbg.hide_debugger()
            dbg.set_callback(EXCEPTION_BREAKPOINT, None)
            self.debugger_attached.set()
        return DBG_CONTINUE

    def prepareBuffer(self, length):
        self.aacdata = create_string_buffer(length)
        self.offset = 0

    def resetTrack(self):
        while not self.bufQueue.empty():
            self.bufQueue.get()
        self.offset = 0

    def getBuffer(self):
        return self.aacdata.raw

    def DoAACFrame(self, dbg):
        # unknown, bufptr, size
        size = dbg.get_arg(2)
        ptr = getDwordAt(dbg, dbg.get_arg(1))
        if size>0 and ptr!=0:
            if self.offset+size>sizeof(self.aacdata):
                debug("Pushing buffer of %d bytes" %self.offset)
                self.bufQueue.put(self.aacdata[:self.offset],block=True)
                self.offset = 0
            copy_process_memory(dbg,ptr,addressof(self.aacdata)+self.offset,size)
            self.offset += size
        return DBG_CONTINUE

    def EndOfTrack (self, dbg=None):
        debug("End of track (bp)")
        if self.onstop:
            self.onstop()
        if self.offset>0:
            debug("Pushing last buffer: %d bytes" %self.offset)
            self.bufQueue.put(self.aacdata[:self.offset],block=True)
            self.offset = 0
        self.bufQueue.put('',block=True)
        return DBG_CONTINUE

    def saveFromQueue(self,f):
        while True:
            buf = self.bufQueue.get(block=True)
            if len(buf):
                debug("Writing buffer of %d bytes" %len(buf))
                f.write(buf)
            else:
                debug("Got end-of-track signal; finishing write thread")
                break

    def patchSndOut(self):
        if self.iTunesVer<len(offsets4):
            self.saveSndOut = self.dbg.read_process_memory(offsets4[self.iTunesVer], 5)     #E8 xx xx xx xx
            self.dbg.write_process_memory(offsets4[self.iTunesVer], '\x33\xC0\x90\x90\x90') #xor eax, eax; nop; nop; nop;

    def restoreSndOut(self):
        if self.iTunesVer<len(offsets4):
            self.dbg.write_process_memory(offsets4[self.iTunesVer], self.saveSndOut)

    def run(self):
        self.dbg = pydbg()
        dbg = self.dbg
        try:
            procname = "iTunes.exe"
            pid = find_pid(dbg, procname)
            if pid!=-1:
                print "attaching to %s(%d)" % (procname, pid)
                dbg.attach(pid)
                #dbg.debug_event_loop()
            else:
                error("process '%s' not found." % procname)

            ver = 0
            found = False
            patched = False
            for ver in xrange(len(offsets1)):
                b1 = dbg.read_process_memory(offsets1[ver], 1)
                if b1=='\x76': #jbe
                    found = True
                    break
                elif b1=='\xEB':
                    patched = True
                    break

            if not found:
                if patched:
                    debug("Already patched")
                    pass
                else:
                    error("Only iTunes %s to %s are supported"%(versions[0],versions[-1]))

            print "Detected iTunes %s" % versions[ver]
            #dbg.write_process_memory(offsets1[ver], '\xEB')
            dbg.set_callback(EXCEPTION_BREAKPOINT, self.handler_breakpoint)
            dbg.bp_set(offsets2[ver], handler=self.DoAACFrame)
            dbg.bp_set(offsets3[ver], handler=self.EndOfTrack)
            self.iTunesVer = ver
            print "Starting debugging"
            #dbg.core_log = lambda msg: sys.stderr.write("[CORE] " + msg + "\n")
            dbg.set_callback(EXCEPTION_ACCESS_VIOLATION, handleAV)
            dbg.run()
        except:
            import traceback
            print >> sys.stderr, "Exception during debugging"
            if debugmode: traceback.print_exc()
            self.stop()
        else:
            #print "Exiting"
            pass

    def stop(self):
        if self.dbg and self.isAlive() and self.debugger_attached.isSet():
            debug("Debugger is alive; stopping it")
            if debugmode:
                self.dbg.core_log = lambda msg: sys.stderr.write("[pydbg] " + msg + "\n")
            #self.dbg.cleanup()
            #self.dbg.set_debugger_active(False)
            self.dbg.detach()
        self.dbg = None

class AACDumper:
    def __init__(self):
        self.p = iTunesController()
        self.dbg = DebugThread()
        self.dbg.start()
        self.dbg.debugger_attached.wait()
        self.dbg.prepareBuffer(2*1024*1024)

    def close(self):
        if self.dbg:
            self.dbg.stop()

    def stop(self):
        debug("Got stop signal; stopping the player")
        self.p.stop()

    def saveThread(self):
        debug("Write thread begin")
        self.mp4.dedrm(self.saveFromQueue)
        self.mp4.save(self.outfilename)
        debug("Write thread end")

    def dumpTrack(self, filename, skip_converted=False):
        #global aacdata, offset
        f1 = os.path.splitext(os.path.abspath(filename))
        if f1[1].lower()!='.m4p':
            print '"%s" is not an m4p file!'%filename
            return
        infile = f1[0]+".m4p"
        outfile = f1[0]+".m4a"
        print 'Processing "%s"' % infile
        if skip_converted and os.path.exists(outfile):
            print '"%s" already exists; skipping conversion.' % outfile
            return
        if not os.path.exists(infile):
            print 'File "%s" does not exist.' % infile
            return
        try:
            self.outfilename = outfile
            self.mp4 = MP4(infile)
            dbg = self.dbg
            dbg.resetTrack()
            print 'Dumping...'
            start = time.time()
            dbg.onstop = self.stop
            self.saveFromQueue = dbg.saveFromQueue
            t = threading.Thread(target=self.saveThread)
            t.start()
            dbg.patchSndOut()
            debug("playTrack begin")
            res = self.p.playTrack(infile)
            debug("playTrack end")
            dbg.onstop = None
            dbg.restoreSndOut()
            if t.isAlive():
                debug("Write thread is still alive; wait for it")
                self.dbg.EndOfTrack()
                t.join()
            if res:
                print "Elapsed: %d seconds" % (time.time()-start)
                print '"%s" saved successfully.' % outfile
            else:
                print "Error while dumping; deleting incomplete output file"
                if os.path.exists(outfile):
                    os.remove(outfile)
            self.mp4 = None
        except MP4Error, e:
            print "Error processing '%s':"%infile
            print "\t",e
        except:
            import traceback
            print >> sys.stderr, "Error while dumping; deleting incomplete output file"
            if os.path.exists(outfile):
                os.remove(outfile)
            if debugmode: traceback.print_exc()

def getMyMusicPath():
    """Get path to the "My Music" folder (C:\Documents and Settings\<username>\My Documents\My Music)"""
    from win32com.shell import shell, shellcon
    try:
        return shell.SHGetFolderPath(0,shellcon.CSIDL_MYMUSIC,0,0)
    except:
        #Win2k/NT4 doesn't support CSIDL_MYMUSIC
        import win32api
        mydocs = shell.SHGetFolderPath(0,shellcon.CSIDL_PERSONAL,0,0)
        h = win32api.RegOpenKey(win32con.HKEY_CURRENT_USER,"SOFTWARE\\Apple Computer, Inc.\\iTunes")
        mymusicName = "My Music"
        try:
            mymusicName = win32api.RegQueryValueEx(h, "Win2KMyMusicFolder")[0]
        finally:
            win32api.RegCloseKey(h)
        return mydocs+"\\"+mymusicName

def convertentity(m):
    v = m.group(2)
    try:
        if m.group(1)=='#':
            if len(v)>0 and v[0]=='x':
                return chr(int(v[1:],16))
            else:
                return chr(int(v))
        return htmlentitydefs.entitydefs[v]
    except (ValueError, KeyError):
        return '&%s;' % v

def converthtml(s):
    return re.sub(r'&(#?)(.+?);',convertentity,s)

class iTunesLibrary:
    def __init__(self,filename=None):
        if not filename:
            mymusic = getMyMusicPath()
            filename = os.path.join(mymusic,"iTunes/iTunes Music Library.xml")
        self.fname = filename
    def fixurl(self,s):
        return unicode(urllib.unquote(converthtml(s)),"utf-8")
    def getM4Plist(self):
        pat = re.compile("<key>Location</key><string>file://localhost/(.*?\.[mM]4[pP])</string>")
        s = file(self.fname,"r").read()
        filelist = pat.findall(s)
        return map(self.fixurl,filelist)

def decode_ansi(s):
    return unicode(s,sys.getfilesystemencoding())
##############
#main code
##############

def main(args):
    print "QTFairUse6 2.4 2006-09-14\nCopyright 2006 Igor Skochinsky <skochinsky@mail.ru>\n"
    if sys.platform != "win32":
        print "This script runs only on Windows!"
        return 1
    if platform.python_version_tuple()[:2] != ['2','4'] or not platform.python_compiler().startswith("MSC"):
        print "Python 2.4 Win32 required!"
        print "Download it from http://www.python.org/download/releases/"
        return 1

    #make sure 'print' doesn't throw an exception if it cannot print character with default encoding...
    writer_class = codecs.getwriter(sys.getfilesystemencoding())
    sys.stdout = writer_class(sys.stdout, 'replace')

    import psyco
    psyco.full()

    skip_converted = False

    if len(args)<2:
        print "No filename specified in the command line. Do you wish to process all protected files in your iTunes library? [Y]"
        choice = raw_input().lower()
        if not choice.startswith("n"):
            l = iTunesLibrary()
            files = l.getM4Plist()
            if len(files)==0:
                print "No protected (.m4p) files found in the library"
                return 0
        else:
            return 0

        choice = raw_input("Skip already converted files (n=overwrite)? [Y]").lower()
        skip_converted = not choice.startswith("n")

    else:
        files = map(decode_ansi,args[1:])

    d = AACDumper() #this will start iTunes if needed
    try:
        n = len(files)
        for i in xrange(n):
            print "=============== Processing file %d/%d ==============="%(i+1,n)
            d.dumpTrack(files[i],skip_converted)
    finally:
        d.close()

args = sys.argv
if "-d" in args:
    debugmode = True
    args.remove("-d")

res = main(args)
raw_input("Press Enter to close\n")
sys.exit(res)
