import sys
try:
    import win32com.client
except ImportError:
    print "win32com is missing!"
    print "Download and install pywin32 from http://sourceforge.net/project/showfiles.php?group_id=78018"
    sys.exit(1)

import win32event, win32api
import pythoncom
import time, threading
from win32com.client import gencache
gencache.EnsureModule('{9E93C96F-CF0D-43F6-8BA8-B807A3370712}', 0, 1, 7)

defaultNamedNotOptArg=pythoncom.Empty

class iTunesEvents:
    def __init__(self):
        self.event = win32event.CreateEvent(None, 0, 0, None)

    def OnPlayerStopEvent(self, iTrack=defaultNamedNotOptArg):
        #thread = win32api.GetCurrentThreadId()
        #print "OnPlayerStopEvent event processed on thread %d"%thread
        if hasattr(self,'callback'):
            self.callback.cbOnStop()
        # Set the event our main thread is waiting on.
        win32event.SetEvent(self.event)

    def OnPlayerPlayEvent(self, iTrack=defaultNamedNotOptArg):
        #thread = win32api.GetCurrentThreadId()
        #print "OnPlayerPlayEvent event processed on thread %d"%thread
        if hasattr(self,'callback'):
            self.callback.cbOnPlay()
        win32event.SetEvent(self.event)

    def OnAboutToPromptUserToQuitEvent(self):
        if hasattr(self,'callback'):
            self.callback.cbOnQuit()
        win32event.SetEvent(self.event)


def WaitWhileProcessingMessages(event, timeout = 2):
    start = time.clock()
    while True:
        # Wake 4 times a second - we can't just specify the
        # full timeout here, as then it would reset for every
        # message we process.
        rc = win32event.MsgWaitForMultipleObjects( (event,), 0,
                                250,
                                win32event.QS_ALLEVENTS)
        if rc == win32event.WAIT_OBJECT_0:
            # event signalled - stop now!
            return True
        if (time.clock() - start) > timeout:
            # Timeout expired.
            return False
        # must be a message.
        pythoncom.PumpWaitingMessages()

class iTunesController:
    def __init__(self, onstart=None, onstop=None, onquit=None):
        self.onstart = onstart
        self.onstop  = onstop
        self.onquit  = onquit
        self.iTunes = win32com.client.Dispatch("iTunes.Application")
        self.iTunes = None

    def stop(self):
        self.cbOnStop()

    def cbOnPlay(self):
        #print "cbOnPlay"
        self.playing = True
        if self.onstart:
            self.onstart()

    def cbOnStop(self):
        #print "cbOnStop"
        if self.playing:
            self.playing = False
            if self.onstop:
                self.onstop()
            self.end = True

    def cbOnQuit(self):
        #print "cbOnQuit"
        self.playing = False
        if self.onquit:
            self.onquit()
        self.end = True

    def playTrack(self, fname):
        self.playing = False
        self.end = False
        try:
            try:
                self.iTunes = win32com.client.DispatchWithEvents("iTunes.Application", iTunesEvents)
                #raise TypeError()
            except:
                print '------------------------\nError connecting to iTunes COM\nTry to reregister iTunes COM by running "iTunes.exe /regserver"\n------------------------'
                self.iTunes = None
                return False

            self.iTunes.callback = self
            self.iTunes.Stop()

            temp_pl_name = "___temp_playlist___"
            temp_pl = self.iTunes.LibrarySource.Playlists.ItemByName(temp_pl_name)
            if temp_pl:
                temp_pl.Delete()
            temp_pl = win32com.client.CastTo(self.iTunes.CreatePlaylist(temp_pl_name),'IITUserPlaylist')
            status = temp_pl.AddFile(fname)
            if not status:
                print 'Failed to play "%s"'%fname
                temp_pl.Delete()
                return False
            while status.InProgress:
                win32api.Sleep(500)
            if status.Tracks.Count==0:
                print '"Failed to play "%s"'%fname
                temp_pl.Delete()
                return False
            t = status.Tracks.Item(1)
            print "Playing: %s - %s"%(t.Artist, t.Name)
            t.Play()

            while not self.end:
                WaitWhileProcessingMessages(self.iTunes.event)

            print "Track ended"
            temp_pl.Delete()
        except:
            import traceback
            print >> sys.stderr, "Exception during playback"
            traceback.print_exc()
            self.iTunes = None
            return False

        return True

if __name__=='__main__':
    def myOnStart():
        print "Track started"
    def myOnStop():
        print "Track stopped"
    p = iTunesController(myOnStart, myOnStop)
    p.playTrack(sys.argv[1])
